!* --------------------------------------------------------------  */
!* (C)Copyright 2006,2008,                                         */
!* International Business Machines Corporation                     */
!* All Rights Reserved.                                            */
!*                                                                 */
!* Redistribution and use in source and binary forms, with or      */
!* without modification, are permitted provided that the           */
!* following conditions are met:                                   */
!*                                                                 */
!* - Redistributions of source code must retain the above copyright*/
!*   notice, this list of conditions and the following disclaimer. */
!*                                                                 */
!* - Redistributions in binary form must reproduce the above       */
!*   copyright notice, this list of conditions and the following   */
!*   disclaimer in the documentation and/or other materials        */
!*   provided with the distribution.                               */
!*                                                                 */
!* - Neither the name of IBM Corporation nor the names of its      */
!*   contributors may be used to endorse or promote products       */
!*   derived from this software without specific prior written     */
!*   permission.                                                   */
!*                                                                 */
!* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
!* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
!* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
!* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
!* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
!* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
!* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
!* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
!* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
!* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
!* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
!* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
!* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
!* --------------------------------------------------------------  */
!* PROLOG END TAG zYx                                              */
!
!
! Example:     3 - dacs_f Cell
! Language:    GNU Fortran
! Core:        PPU
! File:        main_ppu.f90
! Description: Main PPU program - initiates and terminates example,
!              communicates with the SPUs using DaCS Fortran APIs.
!

!
! Main Fortran program that makes use of Fortran array language
! to perform computation and the Fortran bindings for DaCS
! to eliminate separate C programs.
!
!error handler function - registered in main program
      MODULE simpleeh
      CONTAINS
      FUNCTION errhandler (errdata) RESULT(rc)
        include "dacsf.h"
        include "dacsf_interface.h"
        integer(kind=dacs_pvoid_t), intent(in) :: errdata
        integer(kind=dacs_err_t) rc

        integer(kind=dacs_int32_t) :: code
        character(len=256) :: str
        character(len=5) :: shortstr
        integer(kind=dacs_de_id_t) :: de
        integer(kind=dacs_process_id_t) :: pid

        write (*,"(A)") 'error handler invoked'
        call dacsf_error_code(errdata,code,rc)
        call dacsf_error_str(errdata,str,rc)
        call dacsf_error_de(errdata,de,rc)
        call dacsf_error_pid(errdata,pid,rc)

        write(*,"(A,A,A,I8,A,Z8,A,Z16,A)") 'errhandler: str[',trim(str),'] code[',code,'] DE[',de,'] PID[',pid,']'
        rc=0
      END FUNCTION errhandler
      END MODULE simpleeh

      program main
        use simpleeh
        implicit None

        include 'dacsf.h'
        include 'dacsf_interface.h'
        include 'init_cb.h'

        integer, parameter :: N=64  !must match DATA_BUFFER_ENTRIES in control_blk.h
        real :: arr(N)
        integer :: i

        character (len = *), parameter :: pname = "main_ppu: "
        integer(kind=dacs_err_t)   :: rc = 0
        integer(kind=dacs_int32_t) :: count_spes = MAX_UNIT_COUNT
        integer(kind=dacs_int32_t), dimension(1:MAX_UNIT_COUNT) :: rsvd_child_des
        integer(kind=dacs_int64_t), dimension(1:MAX_UNIT_COUNT) :: rsvd_child_pids
        integer(kind=dacs_wid_t),   dimension(1:MAX_UNIT_COUNT) :: ppu_wid
        integer(kind=dacs_int32_t) :: num_rsvd_child
        integer(kind=dacs_int32_t) :: unit_size
        integer(kind=dacs_int32_t) :: unit_count
        integer(kind=dacs_int32_t) :: unit_buffer_size
        integer(kind=dacs_int32_t) :: exit_status=0
        integer(kind=dacs_pvoid_t) :: noparm = 0
        integer(kind=dacs_int32_t) :: flags = 0
        integer(kind=dacs_pvoid_t) :: carr !handle for array address
        integer(kind=dacs_pvoid_t) :: ccb1 !handle for control block address
        integer(kind=dacs_remote_mem_t) ::  carr_remote_mem !handle for array address
        character (len = DACS_MAX_ERRSTR_LEN) :: rc_stg
! separartor for derived types cb1 is a per cent(%) dot(.) is an IBM extension
        type(control_block_t), dimension(1:MAX_UNIT_COUNT) ::  cb1
        integer(kind=dacs_int32_t) :: cb1_element_size
        integer(kind=dacs_int64_t) :: index = 0
        external  main_spu

        character (len = 100) :: buf
        integer(kind= dacs_byte_swap_t) :: ndb = DACS_BYTE_SWAP_DISABLE

        call dacsf_init(DACS_INIT_FLAGS_NONE,rc)
        call errchk('PPE: dacsf_init',rc)

        call dacsf_errhandler_reg(errhandler,flags,rc)
        call errchk('PPE: dacsf_errhandler_reg',rc)

!       initialize the array
        arr = (/(real(i),i=1, N)/)
        write(*,'(A)') 'PPE: initial array value'
        write(*,'(8f6.1)') arr


!       reserve the max spes
        call dacsf_reserve_children(DACS_DE_SPE, count_spes, rsvd_child_des, rc)
        call errchk("PPE: dacsf_reserve_children", rc)
        if (count_spes .lt. 1) then
           call errchk("PPE: Sample needs at least 1 SPE to run.\n",1)
        endif

!       each spe processes equal number of elements which is a multiple of 8
        select case (count_spes)
           case(1)
             unit_size = 64
           case(2:3)
             unit_size = 32
           case(4:7)
             unit_size = 16
           case default
             unit_size = 8
        end select

!       element count cannot be less than minimun in init_cb.h
        unit_size = max(unit_size,MIN_UNIT_SIZE)
        unit_count = DATA_BUFFER_ENTRIES/unit_size !number of spes used in  process group
        unit_buffer_size = unit_size*REAL_SIZE     !size of buffer for a group
        num_rsvd_child = unit_count

!       release extra children
        if (count_spes .gt. num_rsvd_child) then
          call dacsf_release_de_list(count_spes - num_rsvd_child, rsvd_child_des(num_rsvd_child+1:MAX_UNIT_COUNT),rc)
          call errchk("PPE: dacsf_release_de_list", rc)
        endif
        write(*,'(A,I2,A,I2,A,I2,A)') 'array ele count[',n,'] min spu ele count[',MIN_UNIT_SIZE,'] spu ele count[',unit_size,']'
        write(*,'(A,I2,A,I2,A,I2,A)') 'max unit count[', MAX_UNIT_COUNT,'] spus avail[',count_spes,'] spus rsvd[',unit_count,']'

!       create memory to be shared with spes
        call dacsf_makevoid(arr,carr)  !convert array address to handle
        call dacsf_remote_mem_create(carr, DATA_BUFFER_SIZE,DACS_READ_WRITE, carr_remote_mem,rc)
        call errchk("PPE: dacsf_remote_mem_create", rc)

!       initialize the control block for each unit of work
        cb1_element_size = cb_size ! hardcode size in cb
        do i = 1, unit_count
!         load the address into the control block
          cb1(i)%offset = (i-1)*unit_buffer_size
          cb1(i)%chunk_size = unit_buffer_size
          cb1(i)%index = i
        end do


!       allocate SPE tasks
        do i=1, unit_count
          index = cb1(i)%index
!         subroutine name resolves to dacsf_de_start_ptr_embedded
          call dacsf_de_start(rsvd_child_des(i), main_spu, 0_dacs_pvoid_t,0_dacs_pvoid_t,rsvd_child_pids(i),rc)
          call errchk("PPE: dacsf_de_start", rc)

!         initialize a dacs wait identifier for use by dacsf_send
          call dacsf_wid_reserve(ppu_wid(i),rc)
          call errchk("PPE: dacsf_wid_reserve", rc)

!         send the cb
          call dacsf_makevoid(cb1(i),ccb1)  !convert control block address to handle
          call dacsf_send(ccb1, cb1_element_size,rsvd_child_des(i), rsvd_child_pids(i),STREAM_ID,ppu_wid(i),ndb,rc)
          call errchk("PPE: dacsf_send", rc)
        end do

!       wait for each send to complete then share the memory
        do i=1, unit_count
          index = cb1(i)%index
          call dacsf_wait(ppu_wid(i),rc)
          call errchk("PPE: dacsf_wait", rc)

          call dacsf_remote_mem_share(rsvd_child_des(i), rsvd_child_pids(i), carr_remote_mem, rc)
          call errchk("PPE: dacsf_remote_mem_share", rc)
        end do

        write (*,'(A,I2,A)') 'SPEs 0 - ',unit_count,' are now computing'

!       wait for SPEs to all finish
        do i=1, unit_count
          index = cb1(i)%index
          write (*,'(A,I2,A)') 'ppu unit.',index,' waiting on completion'
          call dacsf_de_wait(rsvd_child_des(i), rsvd_child_pids(i), exit_status, rc)
          if (rc .ne. DACS_STS_PROC_FINISHED) then
             call dacsf_strerror(rc,rc_stg)
             write (*,'(A,I4,A,A)') 'dacsf_de_wait rc[',rc,'] - ',trim(rc_stg)
          end if
        end do
!       write out the answer
        write(*,'(A)') 'PPE: final array value'
        write(*,'(8f6.1)') arr

!       cleanup resources
        do i=1, unit_count
          call dacsf_wid_release(ppu_wid(i),rc)
          call errchk("PPE: dacsf_wid_release", rc)
        end do

        call dacsf_remote_mem_destroy(carr_remote_mem,rc)
        call errchk("PPE: dacsf_remote_mem_destroy", rc)


        call dacsf_exit(rc)
      end program

      subroutine errchk(estg,rc)
        include 'dacsf.h'
        character (len=*) :: estg
        integer (kind=dacs_err_t)   :: rc
!       uncomment the following line to see each API call as it returns
!       write(*,"(A,A,I4,A)") estg,' rc[',rc,']'
        if (rc .ne. DACS_SUCCESS) then
          write(*,"(A,A,I4,A)") estg,' rc[',rc,'] program stopped'
          stop
        endif
      end subroutine errchk
