!* --------------------------------------------------------------  */
!* (C)Copyright 2006,2008,                                         */
!* International Business Machines Corporation                     */
!* All Rights Reserved.                                            */
!*                                                                 */
!* Redistribution and use in source and binary forms, with or      */
!* without modification, are permitted provided that the           */
!* following conditions are met:                                   */
!*                                                                 */
!* - Redistributions of source code must retain the above copyright*/
!*   notice, this list of conditions and the following disclaimer. */
!*                                                                 */
!* - Redistributions in binary form must reproduce the above       */
!*   copyright notice, this list of conditions and the following   */
!*   disclaimer in the documentation and/or other materials        */
!*   provided with the distribution.                               */
!*                                                                 */
!* - Neither the name of IBM Corporation nor the names of its      */
!*   contributors may be used to endorse or promote products       */
!*   derived from this software without specific prior written     */
!*   permission.                                                   */
!*                                                                 */
!* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
!* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
!* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
!* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
!* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
!* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
!* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
!* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
!* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
!* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
!* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
!* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
!* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
!* --------------------------------------------------------------  */
!* PROLOG END TAG zYx                                              */
!
!
! Example:     3 - dacs_f Cell
! Language:    GNU Fortran
! Core:        SPU
! File:        main_spu.f90
! Description: Main SPU program - randomizes array elements,
!              communicates with PPU using DaCS Fortran APIs
!

!
! Fortran routine that makes use of Fortran array language
! to perform computation and the Fortran bindings for DaCS
! to eliminate separate C programs
!

      program main_spu
      implicit none

      include 'dacsf.h'
      include 'dacsf_interface.h'
      include 'init_cb.h'

      real, dimension(:), allocatable :: arr
      integer(kind=dacs_err_t) :: rc
      integer(kind=dacs_wid_t) :: spu_wid
      integer(kind=dacs_remote_mem_t) :: spu_remote_mem
      integer :: n,i
      integer(kind=dacs_pvoid_t) :: data

      type(control_block_t)  ::  cb1
      integer(kind=dacs_int32_t) :: cb1_element_size = cb_size !cb size is defined in include file init_cb.h
      real, dimension(:), allocatable :: factr

      call dacsf_init(DACS_INIT_FLAGS_NONE,rc)
      call errchk('SPE: dacsf_init ', rc)

!     initialize a dacs wait identifier
      call dacsf_wid_reserve(spu_wid, rc)
      call errchk('SPE: dacsf_wid_reserve ', rc)

!     receive the control block sent by the ppu
      call dacsf_makevoid(cb1,data)  !convert control block address to handle
      call dacsf_recv(data, cb1_element_size, DACS_DE_PARENT, DACS_PID_PARENT, STREAM_ID, spu_wid, DACS_BYTE_SWAP_DISABLE,rc)
      call errchk('SPE: dacsf_recv ', rc)

      call dacsf_wait(spu_wid,rc)
      call errchk('SPE: dacsf_wait ', rc)

!     accept the array memory shared by the ppu
      call dacsf_remote_mem_accept(DACS_DE_PARENT, DACS_PID_PARENT, spu_remote_mem, rc)
      call errchk('SPE: dacsf_remote_mem_accept ', rc)

!     allocate a work area for the array segment
      n = cb1%chunk_size/REAL_SIZE
      allocate(arr(n))
      allocate(factr(n))

!     copy the array segment to the work area
      call dacsf_makevoid(arr,data)  !convert array address to handle
      call dacsf_get(data, spu_remote_mem, cb1%offset,cb1%chunk_size,spu_wid,DACS_ORDER_ATTR_NONE, DACS_BYTE_SWAP_DISABLE,rc)
      call errchk('SPE: dacsf_get ', rc)

      call dacsf_wait(spu_wid, rc)
      call errchk('SPE: dacsf_wait ', rc)

!     randomize array elements
      call spui_print(' processing array elements_',cb1%index)
      factr = (/ (rand(), i=1, n) /)
      arr(1:n) = arr(1:n) * factr

!     copy the updated array segment to shared memory
      call dacsf_put(spu_remote_mem, cb1%offset, data, cb1%chunk_size,spu_wid, DACS_ORDER_ATTR_NONE, DACS_BYTE_SWAP_DISABLE, rc)
      call errchk('SPE: dacsf_put ', rc)

      call dacsf_wait(spu_wid, rc)
      call errchk('SPE: dacsf_wait ', rc)

!     cleanup resources and terminate
      call dacsf_remote_mem_release(spu_remote_mem,rc)
      call errchk('SPE: dacsf_remote_mem_release ', rc)

      call dacsf_wid_release(spu_wid, rc)
      call errchk('SPE: dacsf_wid_release ', rc)

      call dacsf_exit(rc)
      call errchk('SPE: dacsf_exit ', rc)
      stop
      end

!     If the return code from the API is nonzero, emit a message
!     in a real application further error handling is required.
!
!     estg paramter - the last character in the string is replaced by a NULL in the spu_print.c functions.
      subroutine errchk(estg,rc)
        include 'dacsf.h'
        character (len=*) :: estg
        integer (kind=dacs_err_t)   :: rc
!       uncomment the following line to see each API call as it returns
!       call spu_print32(estg,rc)
        if (rc .ne. DACS_SUCCESS) then
          call spu_print32(estg,rc)
        endif
      end subroutine errchk
