/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//
// Example:     2 - dacs_c Cell
// Language:    IBM XL C
// Core:        SPU
// File:        main_spu.c
// Description: Main SPU program, communicates with PPU using
//              DaCS Fortran APIs, calls Fortran subroutine to
//              randomize array elements
//

#include <stdio.h>
#include <malloc_align.h>
#include <dacs.h>
#include "init_cb.h"
#include "dacs_chk.h"

// local copy of the data array, to be filled by the DMA
float *data;

// local copy of the control block, to be filled by the DMA
control_block cb1 __attribute__ ((aligned(128)));

// Fortran array processing subroutine definition
void compute_f_(float *, int);

//
// Main SPU program which uses DaCS API calls to communcate with the PPU.
// Makes use of Fortran array language by calling a subroutine to process
// the array elements.
//
int main()
{
  DACS_ERR_T rc;
  dacs_wid_t spu_wid;
  dacs_remote_mem_t spu_remote_mem;
  int n;
  int index;

  rc = dacs_init(0);
  ERRCHK("SPE: dacs_init", rc);

  //initialize a dacs wait identifier
  rc = dacs_wid_reserve(&spu_wid);
  ERRCHK("SPE: dacs_wid_reserve", rc);

  //receive the control block sent by the ppu
  rc = dacs_recv(&cb1, sizeof(cb1), DACS_DE_PARENT, DACS_PID_PARENT, STREAM_ID,
                 spu_wid, DACS_BYTE_SWAP_DISABLE);
  ERRCHK("SPE: dacs_recv", rc);

  rc = dacs_wait(spu_wid);
  ERRCHK("SPE: dacs_wait", rc);

  index = cb1.index;
  //accept the array memory shared by the ppu
  rc = dacs_remote_mem_accept(DACS_DE_PARENT, DACS_PID_PARENT,
                              &spu_remote_mem);
  ERRCHK("SPE: dacs_remote_mem_accept", rc);
  //allocate a work area for the array segment
  n = cb1.chunk_size/sizeof(float);
  data = (float *)_malloc_align(cb1.chunk_size, 7);

  //copy the array segment to the work area
  rc = dacs_get(data, spu_remote_mem, cb1.offset,cb1.chunk_size,spu_wid,
                DACS_ORDER_ATTR_NONE, DACS_BYTE_SWAP_DISABLE);
  ERRCHK("SPE: dacs_get", rc);

  rc = dacs_wait(spu_wid);
  ERRCHK("SPE: dacs_wait", rc);

  //invoke the Fortran compute_spu.f90 compute_f subroutine to randomize array elements
  PRINTF("spu unit.%i invoking compute\n",index);
  compute_f_(data, n);

  //copy the updated array segment to shared memory
  rc = dacs_put(spu_remote_mem, cb1.offset, data, cb1.chunk_size,spu_wid,
                DACS_ORDER_ATTR_NONE, DACS_BYTE_SWAP_DISABLE);
  ERRCHK("SPE: dacs_put", rc);

  rc = dacs_wait(spu_wid);
  ERRCHK("SPE: dacs_wait", rc);

  //cleanup resources and terminate
  rc = dacs_remote_mem_release(&spu_remote_mem);
  ERRCHK("SPE: dacs_remote_mem_release", rc);

  rc = dacs_wid_release(&spu_wid);
  ERRCHK("SPE: dacs_wid_release", rc);

  dacs_exit();

  return 0;
}

