/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
//
// Example:     1 - spe2_c Cell
// Language:    IBM XL C
// Core:        SPU
// File:        main_spu.c
// Description: Main SPU program, communicates with PPU using
//              libspe2 APIs, calls Fortran subroutine to randomize
//              array elements
//

#include <spu_mfcio.h>
#include <stdio.h>
#include <spu_intrinsics.h>
#include <malloc_align.h>
#include "init_cb.h"

// local copy of the data array, to be filled by the DMA
float *data;

// local copy of the control block, to be filled by the DMA
control_block cb1 __attribute__ ((aligned(128)));

// Fortran array processing subroutine definition
void compute_f_(float *, int);

//
// Main SPU program which uses libspe2 calls to communcate with the PPU.
// Makes use of Fortran array language by calling a subroutine to process
// the array elements.
//

//the address of the control block is passed as a parameter from the ppu
int main(unsigned long long speid, addr64 argp, addr64 envp)
{
  int n;

  // DMA control block information from system memory - argp must be uint64_t aligned 128
  mfc_get(&cb1, argp.ull, sizeof(cb1), 31, 0, 0);

  mfc_write_tag_mask(1<<31);
  mfc_read_tag_status_all();

  //allocate a work area for the array segment
  n = cb1.chunk_size/sizeof(float);
  data = (float *)_malloc_align(cb1.chunk_size, 7);

  // correctly fetch both 32 and 64 bit address
  uint64_t xaddr = (cb1.addr.ui[1]==0) ? (uint64_t)cb1.addr.ui[0] : cb1.addr.ull;
  //printf("xaddr 0x%016llx\n",xaddr);

  //copy the array segment to the work area
  //DMA the data from system memory to our local store buffer cb1.addr must be uint64_t aligned 128
  mfc_get(data, xaddr, cb1.chunk_size, 31, 0, 0);

  // Wait for the data array DMA to complete.
  mfc_read_tag_status_all();

  //invoke the Fortran compute_spu.f90 compute_f subroutine to randomize array elements
  compute_f_(data, n);

  // DMA the data from local store buffer to system memory
  mfc_put(data, xaddr, cb1.chunk_size, 20, 0, 0);

  mfc_write_tag_mask(1<<20);
  mfc_read_tag_status_all();

  return 0;
}

